﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using System.Web.Routing;
using System.Reflection;
using System.Threading;

namespace AsyncAction.Async
{
    public class AsyncController : Controller
    {
        internal static MethodInfo GetBeginMethod(Type controllerType, string actionName)
        {
            return controllerType.GetMethod("Begin" + actionName,
                BindingFlags.IgnoreCase | BindingFlags.Public | BindingFlags.Instance | BindingFlags.InvokeMethod,
                null, new Type[] { typeof(AsyncCallback), typeof(object) }, null);
        }

        internal static MethodInfo GetEndMethod(Type controllerType, string actionName)
        {
            return controllerType.GetMethod("End" + actionName,
                BindingFlags.IgnoreCase | BindingFlags.Public | BindingFlags.Instance | BindingFlags.InvokeMethod,
                null, new Type[] { typeof(IAsyncResult) }, null);
        }

        private class ActionNotFoundAsyncResult : IAsyncResult
        {
            public ActionNotFoundAsyncResult(object extraData)
            {
                this.AsyncState = extraData;
                this.AsyncWaitHandle = new ManualResetEvent(true);
            }

            #region IAsyncResult Members

            public object AsyncState { get; private set; }

            public WaitHandle AsyncWaitHandle { get; private set; }

            public bool CompletedSynchronously { get { return true; } }

            public bool IsCompleted { get { return true; } }

            #endregion
        }

        private string m_actionName;
        private MethodInfo m_beginMethod;
        private MethodInfo m_endMethod;

        public bool IsAsync { get; private set; }
        
        protected internal virtual IAsyncResult BeginExecute(RequestContext requestContext, AsyncCallback cb, object extraData)
        {
            if (requestContext == null)
            {
                throw new ArgumentNullException("requestContext");
            }

            this.IsAsync = true;

            this.Initialize(requestContext);

            this.TempData.Load(this.ControllerContext, this.TempDataProvider);

            this.m_actionName = this.RouteData.GetRequiredString("action");

            try
            {
                this.m_beginMethod = GetBeginMethod(this.ControllerContext.Controller.GetType(), this.m_actionName);
                this.m_endMethod = GetEndMethod(this.ControllerContext.Controller.GetType(), this.m_actionName);

                if (this.m_beginMethod == null && this.m_endMethod == null)
                {
                    return new ActionNotFoundAsyncResult(extraData);
                }
                else
                {
                    return (IAsyncResult)this.m_beginMethod.Invoke(this.ControllerContext.Controller, new object[] { cb, extraData });                    
                }
            }
            catch
            {
                this.TempData.Save(this.ControllerContext, this.TempDataProvider);
                throw;
            }
        }

        protected internal virtual void EndExecute(IAsyncResult asyncResult)
        {
            try
            {
                if (this.m_beginMethod == null || this.m_endMethod == null)
                {
                    this.HandleUnknownAction(this.m_actionName);
                }
                else
                {
                    var actionResult = (ActionResult)this.m_endMethod.Invoke(this.ControllerContext.Controller, new object[] { asyncResult });
                    actionResult.ExecuteResult(this.ControllerContext);
                }
            }
            finally
            {
                this.TempData.Save(this.ControllerContext, this.TempDataProvider);
            }
        }
    }
}
