﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using System.Web.SessionState;
using System.Web.Routing;

namespace AsyncAction.Async
{
    public class AsyncMvcHandler : IHttpAsyncHandler, IRequiresSessionState
    {
        private ControllerBuilder _controllerBuilder;

        public AsyncMvcHandler(RequestContext requestContext)
        {
            if (requestContext == null)
            {
                throw new ArgumentNullException("requestContext");
            }

            this.Initialize(requestContext);
        }

        protected virtual void Initialize(RequestContext requestContext)
        {
            this.RequestContext = requestContext;

            // Get the controller type
            string controllerName = this.RequestContext.RouteData.GetRequiredString("controller");

            // Instantiate the controller and call Execute
            this.ControllerFactory = this.ControllerBuilder.GetControllerFactory();
            this.Controller = this.ControllerFactory.CreateController(this.RequestContext, controllerName) as AsyncController;
        }

        internal ControllerBuilder ControllerBuilder
        {
            get
            {
                if (_controllerBuilder == null)
                {
                    _controllerBuilder = ControllerBuilder.Current;
                }
                return _controllerBuilder;
            }
            set
            {
                _controllerBuilder = value;
            }
        }

        public RequestContext RequestContext
        {
            get;
            private set;
        }

        internal AsyncController Controller { get; set; }

        internal IControllerFactory ControllerFactory { get; set; }

        #region IHttpAsyncHandler Members

        public IAsyncResult BeginProcessRequest(HttpContext context, AsyncCallback cb, object extraData)
        {
            if (this.Controller == null)
            {
                throw new InvalidOperationException();
            }

            try
            {
                return this.Controller.BeginExecute(this.RequestContext, cb, extraData);
            }
            catch
            {
                this.ControllerFactory.ReleaseController(this.Controller);
                throw;
            }
        }

        public void EndProcessRequest(IAsyncResult result)
        {
            try
            {
                this.Controller.EndExecute(result);
            }
            finally
            {
                this.ControllerFactory.ReleaseController(this.Controller);
            }
        }

        #endregion

        #region IHttpHandler Members

        public bool IsReusable { get { return false; } }

        public void ProcessRequest(HttpContext context)
        {
            throw new NotSupportedException();
        }

        #endregion
    }
}
